'' Driver for Winbond SPI Flash Memory (W25X16AVDAIZ 2MB)
'' This driver also supports 4MB and 8MB flash memories.

'' This is specifically for the TriBlade Prop Blade #1

'' P0  - DIO, SRAM   D0
'' P1  - DO,  SRAM   D1
'' P27 - /CS, SRAM   /CE, Mouse Dta
'' P28 - Clk, EEPROM SCL, Latch LE

'' There's a control block associated with any file transfer.
''
'' ***********************************************************
'' *                    File Name Bytes 1-4                  *
'' ***********************************************************
'' *                    File Name Bytes 5-8                  *
'' ***********************************************************
'' * File Segment *        File Extension Bytes 1-3          *
'' ***********************************************************
'' *    Reserved  *          Current File Address            *
'' ***********************************************************
''
'' File Name      - Up to 8 bytes, padded with $FF
'' File Extension - Up to 3 bytes, padded with $FF
'' File Segment   - $00 to $FE.  $FF used for erased blocks
'' File Address   - Current flash memory address of file position
''
'' Each 4K block has a 16 byte prefix leaving 4080 bytes for data.
'' This prefix contains the first 12 bytes of the file control block,
'' followed by a word with the block number of the next block in the
'' file.  A value of $FFFF indicates that there's no next block.
'' The last word of the prefix is currently unused.
''
'' ***********************************************************
'' *                    File Name Bytes 1-4                  *
'' ***********************************************************
'' *                    File Name Bytes 5-8                  *
'' ***********************************************************
'' * File Segment *        File Extension Bytes 1-3          *
'' ***********************************************************
'' *          Reserved         *      Next Block Number      *
'' ***********************************************************

CON WrData   = $02_000000              ' Write Data / Address / Data
    RdData   = $03_000000              ' Read Data / Address / (Data)
    Status   = $05                     ' Read Status / (Data)
    WrtEna   = $06                     ' Write Enable
    SecEra   = $20_000000              ' Erase Sector / Address
    JEDEC    = $9F                     ' Return JEDEC device info
    JEDEC2M  = $1530EF                 '  2MB Flash Memory (W25X16/16A)
    JEDEC4M  = $1630EF                 '  4MB Flash Memory (W25X32)
    JEDEC8M  = $1730EF                 '  8MB Flash Memory (W25X64)

    Busy     = $01                     ' Busy Status
    AllOnes  = $FFFFFFFF               ' All one bits (erased)

    maxSeg   = $FE                     ' Last file segment (4K)
    secSize  = 4096                    ' Number of bytes in sector
    secHdr   = 16                      ' Space in each sector for header
    secMask  = secSize - 1             ' Mask for sector size
    dataSize = secSize - secHdr        ' Space in sector for user data

VAR long maxSize                       ' Size of Flash Memory
    long CSpin, Clkpin, DIOpin, DOpin  ' I/O pins
    long freeMap[64]                   ' One bit per block (erased blocks)

PUB start(CS, Clk, DIO, DO) | i        ' Initialize the object
   longmove(@CSpin,@CS,4)              ' Save pin numbers
   case sendRecv(JEDEC,3,0)            ' Get the JEDEC code for the device
      JEDEC2M: maxSize := 2 * |<20     ' Winbond W25X16/16A - 2MB Flash
      JEDEC4M: maxSize := 4 * |<20     ' Winbond W25X32     - 4MB Flash
      JEDEC8M: maxSize := 8 * |<20     ' Winbond W25X64     - 8MB Flash
   longfill(@freeMap,0,64)             ' Mark bits of erased blocks
   repeat i from 0 to maxSize-secSize step secSize
      if readData(i+11,0,-1) == $FF    ' Is block erased?
         freeMap[i>>17] |= |< (i>>12 & 31)
         result++                      ' Count free space in blocks

PUB stop

PUB flashSize                          ' Return the size of the memory in bytes
   return maxSize

PUB sendRecv(sD, rCt, rA) | i, m       ' Send and possibly receive
'  sD  = Value to be sent MSB first, right justified in parameter
'  rCt = Number of bytes to be transferred (>0 received, <0 sent)
'  rA  = 0 or address of buffer area.  If 0, @RESULT is used
   outa[DIOpin]~
   dira[DIOpin]~~                      ' Set up the various I/O pins
   dira[DOpin]~
   outa[Clkpin]~
   dira[Clkpin]~~
   outa[CSpin]~
   dira[CSpin]~~
   if rA == 0                          ' Use RESULT if rA is zero
      rA := @result
   m := $FF000000
   repeat until m == 0 or sD & m <> 0  ' Set up sending mask
      m >>= 8
   m &= $80808080                      ' Start with most significant bit
   repeat until m == 0
      outa[DIOpin] := (sD & m) <> 0    ' Output bit to be sent
      outa[Clkpin]~~                   ' Toggle clock pulse
      outa[Clkpin]~
      m >>= 1                          ' Advance mask towards LSB
   if rCt < 0
      repeat i from 0 to -rCt-1
         m := $80                      ' Set up sending mask
         sD := byte[rA+i]              '  and get data byte
         repeat until m == 0
            outa[DIOpin] := (sD & m) <> 0
            outa[Clkpin]~~             ' Output bit to be set
            outa[Clkpin]~              ' Toggle clock pulse
            m >>= 1                    ' Advance mask towards LSB
   dira[DIOpin]~                       ' Don't need DIO pin now
   if rCt > 0
      repeat i from 0 to rCt-1
         m := $80                      ' Set up receiving mask
         sD~                           ' Make sure byte is zeroed
         repeat until m == 0
            if ina[DOpin] <> 0         ' Input bit to be received
               sD |= m
            outa[Clkpin]~~             ' Toggle clock pulse
            outa[Clkpin]~
            m >>= 1                    ' Advance mask towards LSB
         byte[rA+i] := sD
   outa[CSpin]~~                       ' All done, deselect device
   dira[DOpin]~
   dira[Clkpin]~                       ' Release the various I/O pins
   dira[CSpin]~

PUB readData(addr, data, count)        ' Read bytes from flash memory
'  addr  = Flash memory starting address for reading data
'  data  = Hub starting address for data
'  count = Number of bytes to be read from flash memory
'          If count is -1 to -4, use result for data area
   if count < 0
      sendRecv(RdData|addr,-count,@result)
   else
      sendRecv(RdData|addr,count,data)

PUB writeData(addr, data, count) | p, n ' Write bytes to flash memory
'  addr  = Flash memory starting address for writing data
'  data  = Hub starting address for data or data itself
'  count = Number of bytes to be written to flash memory
'          If count is -1 to -4, data itself is in data
'  This routine handles the process of writing to flash memory
'  in pages of 256 bytes or less (aligned to 256 page boundaries)
   if count < 0
      p := @data
      count := -count
   else
      p := data
   repeat while count > 0              ' Handle end of page, full pages,
      n := count <# (256 - (addr&$FF)) '  and last partial page
      sendRecv(WrtEna,0,0)             ' Enable writes
      sendRecv(WrData|addr,-n,p)
      repeat until sendRecv(Status,1,0) & Busy == 0 ' Wait until done
      addr += n
      p += n                           ' Advance pointers
      count -= n

PUB eraseData(addr)                    ' Erase a 4K sector of flash
'  addr  = Flash memory address within sector to be erased
   sendRecv(WrtEna,0,0)                ' Enable writes
   sendRecv(SecEra|addr,0,0)           ' Erase 4K sector
   freeMap[addr>>17] |= |< (addr>>12 & 31) ' Mark sector
   repeat until sendRecv(Status,1,0) & Busy == 0 ' Wait until done

PUB initFile(p,s) | i, c               ' Initialize file control block
'  p     = Address of file control block as described earlier
'  s     = Address of start of a string containing a file name
'  returns true if a valid file name (and extension) is parsed, false otherwise
'          The file control block is initialized to the normalized file name
   repeat i from 0 to 3
      long[p][i] := AllOnes
   i := 0
   repeat
      case c := byte[s++]
         "-", "0".."9", "A".."Z", "_", "a".."z":
            if i < 8
               byte[p][i++] := c
         ".":
            quit
         0: return true
         other:
            return false
   i := 8
   repeat
      case c := byte[s++]
         "-", "0".."9", "A".."Z", "_", "a".."z":
            if i < 11
               byte[p][i++] := c
         0: return true
         other:
            return false

PUB findFile(p) | i,x,z,a,b,c          ' Find match for specified file
'  p     = Address of file control block as described earlier
'  returns true if found, false if not found.  File control block updated.
'  The file control block is used to choose the starting segment for the
'  search and the "random" operator is applied as well to attempt to
'  scatter the file segments around the flash memory to provide some
'  "wear levelling".
   x := long[p][0] ^ long[p][1] ^ long[p][2]
   x := x? & constant(!secMask)
   repeat i from 0 to maxSize-secSize step secSize
      z := (i + x) & (maxSize-1)         ' Scatter file segments around
      if freeMap[z>>17] & |< (z>>12 & 31) ' Check for free blocks
         next
      readData(z,@a,12)
      if long[p][0] == a and long[p][1] == b and long[p][2] == c
         long[p][3] := byte[p][15] << 24 | (z + 16)
         return true
   return false

PUB findFree(p) | i,x,z,a,b,c          ' Find erased file segment
'  p     = Address of file control block as described earlier
'  returns true if found, false if not found.  File control block updated.
'  The file control block is used to choose the starting segment for the
'  search and the "random" operator is applied as well to attempt to
'  scatter the file segments around the flash memory to provide some
'  "wear levelling".  The first erased segment found is initialized.
   x := long[p][0] ^ long[p][1] ^ long[p][2]
   x := x? & constant(!secMask)
   repeat i from 0 to maxSize-secSize step secSize
      z := (i + x) & (maxSize-1)       ' Scatter file segments around
      ifnot freeMap[z>>17] & |< (z>>12 & 31) ' Check for used blocks
         next
      readData(z,@a,12)
      if a == AllOnes and b == AllOnes and c == AllOnes
         long[p][3] := byte[p][15] << 24 | (z + 16)
         writeData(z,p,12)             ' Write file name & segment #
         freeMap[z>>17] &= ! |< (z>>12 & 31) ' Mark as not erased
         return true
   return false

PUB firstFile(p)                      ' Init file control block for dir list
   long[p][3] := byte[p][15] << 24    ' Start with first 4K segment in memory

PUB nextFile(p) | t                   ' Look for next file in flash memory
   repeat t from long[p][3] & $00FFFFFF to maxSize-secSize step secSize
      if freeMap[t>>17] & |< (t>>12 & 31) ' Check for erased block
         next
      readData(t,p,12)                ' Get the directory entry
      if byte[p][11] == 0             ' Is this the 1st segment in the file?
         long[p][3] := byte[p][15] << 24 | (t + secSize)
         return true
   return false

PUB openFile(p)                       ' Open file for reading
'  p     = Address of file control block as described earlier
'  returns true if successful, false if file doesn't exist.
'          File control block initialized for first read from start of file
   long[p][2] &= $00FFFFFF            ' Set current segment to $00
   return findFile(p)                 ' Look for specified file

PUB createFile(p) | s                 ' Open file for writing
'  p     = Address of file control block as described earlier
'  returns true if successful, false if file currently exists already.
'          File control block initialized for first write to start of file
   long[p][2] &= $00FFFFFF            ' Set current segment to $00
   ifnot findFile(p)                  ' Specified file must not be there
      if findFree(p)                  ' If absent, allocate first segment
         return true
   return false

PUB eraseFile(p) | c, n               ' Erase existing file
'  p     = Address of file control block as described earlier
'  returns true if successful, false if file doesn't exist
'          All blocks / segments of the file are erased.  The file control
'          block is left so that createFile can be called immediately after
   long[p][2] &= $00FFFFFF            ' Set current segment to $00
   if findFile(p)                     ' Specified file must be present
      c := long[p][3] & $00FFF000     ' Start of current file segment
      repeat until c == $0FFFF000     ' Continue until link == $FFFF
         n := c
         c := readData(n+12,0,-2)<<12 ' Get link to next segment
         eraseData(n)                 ' Erase current segment
      return true
   return false

PUB readFile(p,a,c) | f, n, h         ' Read data from a file
'  p     = Address of file control block as described earlier
'  a     = Address of area to receive data from the file
'  c     = Count of number of bytes to be read
'  returns true if successful, false if an end of file occurs
'          The read begins at the Current File Address in the file control
'          block.  The Current File Address is updated as necessary.  Note
'          that there is no end of file marker other than the lack of further
'          4K blocks in the file.
   f := long[p][3] & $00FFFFFF        ' Current file position
   h := f & $FFF000 + 12              ' Position of link to next block
   if f & $FFF => secHdr              ' Positioned in data area?
      n := c <# $1000 - (f & $FFF)    ' Remainder in block
      readData(f,a,n)                 ' Read the data
      f += n
      if f & $FFF == 0                ' File position at end of block?
         f -= secSize                 ' Leave pointing at beginning for
      a += n                          '  access to the next block link
      c -= n                          ' Advance the pointers
   repeat while c > 0                 ' Any blocks left to read?
      if byte[p][11] == $FE           ' Too many segments?
         return false
      byte[p][11] += 1                ' Increment segment number
      f := readData(h,0,-2)<<12       ' Get link to next block
      if f == $0FFFF000               ' Reached end of file
         return false
      h := f & $FFF000 + 12           ' Position of link to next block
      f += secHdr                     ' Start of data area
      n := c <# dataSize              ' Amount to read
      readData(f,a,n)                 ' Read the data
      f += n
      if f & $FFF == 0                ' File position at end of block?
         f -= secSize                 ' Leave pointing at beginning for
      a += n                          '  access to the next block link
      c -= n                          ' Advance the pointers
   long[p][3] := byte[p][15] << 24 | f ' Update file pointer
   return true

PUB writeFile(p,a,c) | f, n, h        ' Write data to a file
'  p     = Address of file control block as described earlier
'  a     = Address of area containing data to be written to the file
'  c     = Count of number of bytes to be written
'  returns true if successful, false if new block needed and none available
'          The write begins at the Current File Address in the file control
'          block.  The Current File Address is updated as necessary.  When
'          the write operation fills the current block, a new block is
'          allocated, the block's header is initialized, and the write
'          operation continues in the new block.
   f := long[p][3] & $00FFFFFF        ' Current file position
   h := f & $FFF000 + 12              ' Position of link to next block
   if f & $FFF => secHdr              ' Positioned in data area?
      n := c <# $1000 - (f & $FFF)    ' Remainder in block
      writeData(f,a,n)                ' Write the data
      f += n
      if f & $FFF == 0                ' File position at end of block?
         f -= secSize                 ' Leave pointing at beginning for
      a += n                          '  access to the next block link
      c -= n                          ' Advance the pointers
   repeat while c > 0                 ' Any blocks left to write?
      if byte[p][11] == $FE           ' Too many segments?
         return false
      byte[p][11] += 1                ' Increment segment number
      ifnot findFree(p)               ' Request a new one
         return false
      f := long[p][3] & $00FFFFFF
      writeData(h,f>>12,-2)           ' Update file link
      h := f & $FFF000 + 12           ' Position of link to next block
      n := c <# dataSize              ' Amount to write
      writeData(f,a,n)                ' Write the data
      f += n
      if f & $FFF == 0                ' File position at end of block?
         f -= secSize                 ' Leave pointing at beginning for
      a += n                          '  access to the next block link
      c -= n                          ' Advance the pointers
   long[p][3] := byte[p][15] << 24 | f ' Update file pointer
   return true
